#include <glew.h>
#include <glut.h>
#include "FlyingCamera.h"
#include "BasicModel.h"
#include "Shader.h"
#include "DeferredShadingEngine.h"
#include "ScreenDraw.h"

/* #DEFINES */

#define CAMERA_SPEED 0.01f //speed of camera as a float
#define INVERT 1 //true or false
#define MOUSE_SENS 0.5 //float between 0.0 and 1.0

/* VARIABLES */

int width, height, curWidth, curHeight;
FlyingCamera* camera;
bool keysDown[6];
int mouseX, mouseY;
const int NUM_MODELS = 3;
BasicModel* bm[NUM_MODELS];
int model;
char renderType[100]; //what type of rendering is currently displaying
//MRT/Deferred things
DeferredShadingEngine* dfe;
enum    RENDER_MODE{FIXED_FUNCTION, DIRECTIONAL, POINT, POINT_W_LIGHT_GEOM, FULL_DEFERRED, FULL_DEFERRED_W_LIGHT_GEOM, NORMALS, DEPTH, COLORS, NUM_MODES};
int     mode;
// vars for fps
int frame, elapsedTime, timebase, timeNow, timeLast;
float fps;

/* FUNCTIONS */

void display();
void useFixedFunction();
void useDeferred();
void doDirectionalAndAmbient();
void doPoint();
void drawPointLights();
void drawLittleLights();
void doDirectTarget(GLuint, bool=false);
void drawFPS();
void drawCamPos();
void drawScene();
void view();
void lights();
void geometry();
void idle();
void reshape(int, int);
void keyboardDown(unsigned char, int, int);
void keyboardUp(unsigned char, int, int);
void motion(int, int);
void passiveMotion(int, int);

/* REAL CODE */

void display() {
   timeNow = glutGet(GLUT_ELAPSED_TIME);
   //handle movement
   if( keysDown[0] ) //w
      camera->moveForward(CAMERA_SPEED * (timeNow - timeLast));
   if( keysDown[1] ) //a
      camera->moveLeft(CAMERA_SPEED * (timeNow - timeLast));
   if( keysDown[2] ) //s
      camera->moveBackward(CAMERA_SPEED * (timeNow - timeLast));
   if( keysDown[3] ) //d
      camera->moveRight(CAMERA_SPEED * (timeNow - timeLast));

   //DRAW w/ mode switch
   glClearColor(0.0, 0.0, 0.0, 1.0);
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   if( mode%NUM_MODES == FIXED_FUNCTION ) {
      drawScene();
      sprintf_s(renderType,"Rendering: OpenGL");
   }
   else {
      useDeferred();
   }

   //SCREEN TEXT
   glColor3f(1.0,0.0,0.0); //red text
   glDisable(GL_DEPTH_TEST);
   drawFPS();
   drawCamPos();
   drawText(5,95, GLUT_BITMAP_HELVETICA_12, renderType);
   glEnable(GL_DEPTH_TEST);

   /* ACTION */
   glutSwapBuffers();
   glutPostRedisplay();
   timeLast = timeNow;
}

void useDeferred() {

   //BUILD G-BUFFER
   dfe->preDrawScene();
   drawScene();
   dfe->postDrawScene();

   //RENDER SCENE USING G-BUFFER
   if( mode%NUM_MODES == NORMALS ) {
      dfe->drawNormals(curWidth, curHeight);
      sprintf_s(renderType,"Rendering: Normals");
   }
   else if( mode%NUM_MODES == DEPTH ) {
      dfe->drawDepth(curWidth, curHeight, camera->zNear, camera->zFar);
      sprintf_s(renderType,"Rendering: Depth Map");
   }
   else if( mode%NUM_MODES == COLORS ) {
      dfe->drawColors(curWidth, curHeight);
      sprintf_s(renderType,"Rendering: Diffuse Material Colors");
   }
   else if( mode%NUM_MODES == DIRECTIONAL ) {
      dfe->drawDirectionalAndAmbient(curWidth, curHeight);
      sprintf_s(renderType,"Rendering: Directional Lighting");
   }
   else if( mode%NUM_MODES == POINT || mode%NUM_MODES == POINT_W_LIGHT_GEOM ) {
      dfe->preDrawPointLights(curWidth, curHeight, camera->zNear, camera->zFar);
      drawPointLights();
      dfe->postDrawPointLights();
      sprintf_s(renderType,"Rendering: Point Lighting");
      //draw point lights for reference
      if(mode%NUM_MODES == POINT_W_LIGHT_GEOM) {
         glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
         glDisable(GL_CULL_FACE);
         glColor3f(0.5,0,0);
         drawPointLights();
         glEnable(GL_CULL_FACE);
         glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
         sprintf_s(renderType,"Rendering: Point Lighting with Light Geometry");
      }
      drawLittleLights();
   }
   else if( mode%NUM_MODES == FULL_DEFERRED || mode%NUM_MODES == FULL_DEFERRED_W_LIGHT_GEOM ) {
      dfe->drawDirectionalAndAmbient(curWidth, curHeight);
      dfe->preDrawPointLights(curWidth, curHeight, camera->zNear, camera->zFar);
      drawPointLights();
      dfe->postDrawPointLights();
      sprintf_s(renderType,"Rendering: Full Deferred Lighting");
      //draw point lights for reference
      if(mode%NUM_MODES == FULL_DEFERRED_W_LIGHT_GEOM) {
         glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
         glDisable(GL_CULL_FACE);
         glColor3f(0.5,0,0);
         drawPointLights();
         glEnable(GL_CULL_FACE);
         glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
         sprintf_s(renderType,"Rendering: Full Deferred Lighting with Light Geometry");
      }
      drawLittleLights();
   }
}

void drawPointLights() {
   glDisable(GL_LIGHTING);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   view();
   //geometry()

      //top

      dfe->setPointLightColor(1.0, 1.0, 1.0);
      glPushMatrix();
      glTranslatef(0, 1.0, 0);
      glutSolidSphere(2, 10, 10);
      glPopMatrix();

      //back left clockwise

      dfe->setPointLightColor(1.0, 0.0, 0.0);
      glPushMatrix();
      glTranslatef(-2.5, 1.0, -2.5);
      glutSolidSphere(2, 10, 10);
      glPopMatrix();

      dfe->setPointLightColor(1.0, 1.0, 1.0);
      glPushMatrix();
      glTranslatef(-1.25, 1.0, -2.75);
      glutSolidSphere(2, 10, 10);
      glPopMatrix();

      dfe->setPointLightColor(0.0, 1.0, 0.0);
      glPushMatrix();
      glTranslatef(0.0, 0.5, -3);
      glutSolidSphere(2, 10, 10);
      glPopMatrix();

      dfe->setPointLightColor(1.0, 1.0, 1.0);
      glPushMatrix();
      glTranslatef(1.25, 1.0, -2.75);
      glutSolidSphere(2, 10, 10);
      glPopMatrix();

      dfe->setPointLightColor(0.0, 0.0, 1.0);
      glPushMatrix();
      glTranslatef(2.5, 1.0, -2.5);
      glutSolidSphere(2, 10, 10);
      glPopMatrix();

      dfe->setPointLightColor(1.0, 1.0, 1.0);
      glPushMatrix();
      glTranslatef(2.75, 1.0, -1.25);
      glutSolidSphere(2, 10, 10);
      glPopMatrix();

      dfe->setPointLightColor(1.0, 1.0, 0.0);
      glPushMatrix();
      glTranslatef(3, 0.5, 0);
      glutSolidSphere(2, 10, 10);
      glPopMatrix();

      dfe->setPointLightColor(1.0, 1.0, 1.0);
      glPushMatrix();
      glTranslatef(2.75, 1.0, 1.25);
      glutSolidSphere(2, 10, 10);
      glPopMatrix();

      dfe->setPointLightColor(1.0, 0.0, 1.0);
      glPushMatrix();
      glTranslatef(2.5, 1.0, 2.5);
      glutSolidSphere(2, 10, 10);
      glPopMatrix();

      dfe->setPointLightColor(1.0, 1.0, 1.0);
      glPushMatrix();
      glTranslatef(1.25, 1.0, 2.75);
      glutSolidSphere(2, 10, 10);
      glPopMatrix();

      dfe->setPointLightColor(1.0, 0.0, 1.0);
      glPushMatrix();
      glTranslatef(0.0, 0.5, 3);
      glutSolidSphere(2, 10, 10);
      glPopMatrix();

      dfe->setPointLightColor(1.0, 1.0, 1.0);
      glPushMatrix();
      glTranslatef(-1.25, 1.0, 2.75);
      glutSolidSphere(2, 10, 10);
      glPopMatrix();
      
      dfe->setPointLightColor(1.0, 1.0, 0.0);
      glPushMatrix();
      glTranslatef(-2.5, 1.0, 2.5);
      glutSolidSphere(2, 10, 10);
      glPopMatrix();

      dfe->setPointLightColor(1.0, 1.0, 1.0);
      glPushMatrix();
      glTranslatef(-2.75, 1.0, 1.25);
      glutSolidSphere(2, 10, 10);
      glPopMatrix();

      dfe->setPointLightColor(0.0, 0.0, 1.0);
      glPushMatrix();
      glTranslatef(-3, 0.5, 0);
      glutSolidSphere(2, 10, 10);
      glPopMatrix();

      dfe->setPointLightColor(1.0, 1.0, 1.0);
      glPushMatrix();
      glTranslatef(-2.75, 1.0, -1.25);
      glutSolidSphere(2, 10, 10);
      glPopMatrix();

}

void drawLittleLights() {
   glDisable(GL_LIGHTING);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   view();
   //geometry()

      //top

      glColor3f(1.0, 1.0, 1.0);
      glPushMatrix();
      glTranslatef(0, 1.0, 0);
      glutSolidSphere(0.1, 10, 10);
      glPopMatrix();

      //back left clockwise

      glColor3f(1.0, 0.0, 0.0);
      glPushMatrix();
      glTranslatef(-2.5, 1.0, -2.5);
      glutSolidSphere(0.1, 10, 10);
      glPopMatrix();

      glColor3f(1.0, 1.0, 1.0);
      glPushMatrix();
      glTranslatef(-1.25, 1.0, -2.75);
      glutSolidSphere(0.1, 10, 10);
      glPopMatrix();

      glColor3f(0.0, 1.0, 0.0);
      glPushMatrix();
      glTranslatef(0.0, 0.5, -3);
      glutSolidSphere(0.1, 10, 10);
      glPopMatrix();

      glColor3f(1.0, 1.0, 1.0);
      glPushMatrix();
      glTranslatef(1.25, 1.0, -2.75);
      glutSolidSphere(0.1, 10, 10);
      glPopMatrix();

      glColor3f(0.0, 0.0, 1.0);
      glPushMatrix();
      glTranslatef(2.5, 1.0, -2.5);
      glutSolidSphere(0.1, 10, 10);
      glPopMatrix();

      glColor3f(1.0, 1.0, 1.0);
      glPushMatrix();
      glTranslatef(2.75, 1.0, -1.25);
      glutSolidSphere(0.1, 10, 10);
      glPopMatrix();

      glColor3f(1.0, 1.0, 0.0);
      glPushMatrix();
      glTranslatef(3, 0.5, 0);
      glutSolidSphere(0.1, 10, 10);
      glPopMatrix();

      glColor3f(1.0, 1.0, 1.0);
      glPushMatrix();
      glTranslatef(2.75, 1.0, 1.25);
      glutSolidSphere(0.1, 10, 10);
      glPopMatrix();

      glColor3f(1.0, 0.0, 1.0);
      glPushMatrix();
      glTranslatef(2.5, 1.0, 2.5);
      glutSolidSphere(0.1, 10, 10);
      glPopMatrix();

      glColor3f(1.0, 1.0, 1.0);
      glPushMatrix();
      glTranslatef(1.25, 1.0, 2.75);
      glutSolidSphere(0.1, 10, 10);
      glPopMatrix();

      glColor3f(1.0, 0.0, 1.0);
      glPushMatrix();
      glTranslatef(0.0, 0.5, 3);
      glutSolidSphere(0.1, 10, 10);
      glPopMatrix();

      glColor3f(1.0, 1.0, 1.0);
      glPushMatrix();
      glTranslatef(-1.25, 1.0, 2.75);
      glutSolidSphere(0.1, 10, 10);
      glPopMatrix();
      
      glColor3f(1.0, 1.0, 0.0);
      glPushMatrix();
      glTranslatef(-2.5, 1.0, 2.5);
      glutSolidSphere(0.1, 10, 10);
      glPopMatrix();

      glColor3f(1.0, 1.0, 1.0);
      glPushMatrix();
      glTranslatef(-2.75, 1.0, 1.25);
      glutSolidSphere(0.1, 10, 10);
      glPopMatrix();

      glColor3f(0.0, 0.0, 1.0);
      glPushMatrix();
      glTranslatef(-3, 0.5, 0);
      glutSolidSphere(0.1, 10, 10);
      glPopMatrix();

      glColor3f(1.0, 1.0, 1.0);
      glPushMatrix();
      glTranslatef(-2.75, 1.0, -1.25);
      glutSolidSphere(0.1, 10, 10);
      glPopMatrix();

}

void drawFPS() {
   //calc fps
   frame++;
   if( timeNow - timebase > 1000 ) { //Only refresh every second
      fps = frame*1000.0/(timeNow-timebase);
      timebase = timeNow;
      frame = 0;
   }
   //write fps
   char fpsString[30];
   sprintf_s(fpsString, "fps: %g", fps);
   //draw fps
   glColor3f(1.0,0.0,0.0);
   drawText(85,95, GLUT_BITMAP_HELVETICA_12, fpsString);
}

void drawCamPos() {
   char posString[30];
   sprintf_s(posString, "<%.2f, %.2f, %.2f>", camera->loc.x, camera->loc.y, camera->loc.z);
   glColor3f(1.0,0.0,0.0);
   drawText(80,90, GLUT_BITMAP_HELVETICA_12, posString);
}

void drawScene() {
   //draw
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   view();
   lights();
   geometry();
}

void view() {
   /* CAMERA */
   gluLookAt(  camera->loc.x, camera->loc.y, camera->loc.z, // eye location
               camera->loc.x+camera->dir.x, camera->loc.y+camera->dir.y, camera->loc.z+camera->dir.z, // Look at point
               camera->up.x, camera->up.y, camera->up.z // up vector
            );
}

void lights() {
   /* LIGHTS */
   glEnable(GL_LIGHTING);
   glEnable(GL_LIGHT0);
   //Point light position
   GLfloat lightPos[] = {0.0, 5.0, 0.0, 1.0};
   //Ambient composition
   GLfloat ambientComp[] = {0.0, 0.0, 0.0, 0.0};
   //Diffuse composition
   GLfloat diffuseComp[] = {0.5, 0.5, 0.5, 0.0};
   //Specular composition
   GLfloat specularComp[] = {0.0, 0.0, 0.0, 0.0};
   //Ambient light
   GLfloat globalAmbient[] = {0.2, 0.2, 0.2, 0.0};
   glLightModelfv(GL_LIGHT_MODEL_AMBIENT, globalAmbient);
   //Point light
   glLightfv(GL_LIGHT0, GL_POSITION, lightPos);
   glLightfv(GL_LIGHT0, GL_AMBIENT, ambientComp);
   glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuseComp);
   glLightfv(GL_LIGHT0, GL_SPECULAR, specularComp);
}

void geometry() {
   /* DRAW GEOMETRY */

   // Draw Ground Plane
   glColor3f(1.0,1.0,1.0);
   GLfloat groundMat[] = {0.7, 0.6, 0.3, 1.0};
   glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, groundMat);
   GLfloat spec[] = {0.0,0.0,0.0,1.0};
   glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, spec);
   glBegin(GL_QUADS);
      glNormal3f(0.f, 200.f, 0.f);
      glVertex3f(-5,0,-5);
      glVertex3f(-5,0,5);
      glVertex3f(5,0,5);
      glVertex3f(5,0,-5);
   glEnd();


   //diffuse and ambient material properties
   GLfloat matColor[4] = {1.0f, 1.0f, 1.0f, 1.0f};
   //specular
   GLfloat matSpec[4] = {1.0f, 1.0f, 1.0f, 1.0f};
   GLfloat shininess = 128.0f;

   glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, matColor);
   glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, matSpec);
   glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, shininess);

   glPushMatrix();
   glTranslatef(0,0.5,0);
   bm[model]->draw();
   glPopMatrix();
}

void idle() {
   glutPostRedisplay();
}

void reshape(int w, int h) {
   curWidth = w;
   curHeight = h;
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(camera->fov,
                  float(w)/h,
                  camera->zNear,
                  camera->zFar);
   glMatrixMode(GL_MODELVIEW);
   glViewport(0, 0, w, h);
}

void keyboardDown(unsigned char key, int x, int y ) {
   switch( key ) {
      case 'w':
         keysDown[0] = true;
         break;
      case 'a':
         keysDown[1] = true;
         break;
      case 's':
         keysDown[2] = true;
         break;
      case 'd':
         keysDown[3] = true;
         break;
   }
}

void keyboardUp(unsigned char key, int x, int y ) {
   switch( key ) {
      case 'q':
         mode += NUM_MODES-1;
         break;
      case 'w':
         keysDown[0] = false;
         break;
      case 'a':
         keysDown[1] = false;
         break;
      case 's':
         keysDown[2] = false;
         break;
      case 'd':
         keysDown[3] = false;
         break;
      case 'e':
         ++mode;
         break;
      case '1':
         model = 0;
         break;
      case '2':
         model = 1;
         break;
      case '3':
         model = 2;
         break;
   }
}

void motion(int x, int y) {
   //Update camera
   camera->setYaw(float(x-mouseX)*MOUSE_SENS);
   float dy;
   if(INVERT)
      dy = mouseY-y;
   else
      dy = y-mouseY;
   camera->setPitch(dy*MOUSE_SENS);

   //Store new mouse position
   mouseX = x;
   mouseY = y;

   //Display changes
   glutPostRedisplay();
}

void passiveMotion(int x, int y) {
   mouseX = x;
   mouseY = y;
}

int main(int argc, char** argv) {

   width = curWidth = 800;
   height = curHeight = 600;
   mode = FULL_DEFERRED;
   elapsedTime = frame = timebase = timeNow = timeLast = 0;
   model = 0;

   // Create the camera
   camera = new FlyingCamera(45.0,       // fov
                             0, 5, 5,   // location
                             1.0, 100.0);// zNear, zFar
   camera->setPitch(-45);

   //Load model
   bm[0] = new BasicModel("./assets/bunny500.m");
   bm[1] = new BasicModel("./assets/gargoyle_2k.m");
   bm[2] = new BasicModel("./assets/dragon10k.m");

   //OpenGL setup
   glutInit(&argc, argv);
   glutInitWindowSize(width, height);
   glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
   glutCreateWindow("Deferred Shading");
   glEnable(GL_DEPTH_TEST);
   glDepthFunc(GL_LESS);
   glEnable(GL_NORMALIZE);
   glEnable(GL_CULL_FACE);
   glCullFace(GL_BACK);

   //register callback functions
   glutDisplayFunc(display);
   glutIdleFunc(idle);
   glutReshapeFunc(reshape);
   glutKeyboardFunc(keyboardDown);
   glutKeyboardUpFunc(keyboardUp);
   glutMotionFunc(motion);
   glutPassiveMotionFunc(passiveMotion);

   //Error check
   GLenum err = glewInit();
   if (err != GLEW_OK || !glewIsSupported("GL_VERSION_2_0"))
   {
      printf("OpenGL 2.0 not supported. No shaders!\n");
      printf("%s\n", glewGetErrorString(err));
      printf("%s\n", (char*)glGetString( GL_VERSION ) );
      exit(-1);
   }
   printf("OpenGL 2.0 supported. Using OpenGL %s \n\n",(char*)glGetString( GL_VERSION ));

   //engine
   dfe = new DeferredShadingEngine(width, height);
   dfe->init();

   //start the main loop
   glutMainLoop();
}